@@ -41,10 +41,11 @@
 
 #include <argmatch.h>
 #include <closeout.h>
+#include <configmake.h>
 #include <exitfail.h>
 #include <getdate.h>
-#include <localedir.h>
 #include <rmt.h>
+#include <rmt-command.h>
 #include <prepargs.h>
 #include <quotearg.h>
 #include <version-etc.h>
@@ -254,6 +255,7 @@ enum
   DELETE_OPTION,
   EXCLUDE_CACHES_OPTION,
   EXCLUDE_OPTION,
+  EXCLUDE_TAG_OPTION,
   FORCE_LOCAL_OPTION,
   GROUP_OPTION,
   HANG_OPTION,
@@ -603,6 +605,8 @@ static struct argp_option options[] = {
    N_("exclude patterns listed in FILE"), GRID+1 },
   {"exclude-caches", EXCLUDE_CACHES_OPTION, 0, 0,
    N_("exclude directories containing a cache tag"), GRID+1 },
+  {"exclude-tag", EXCLUDE_TAG_OPTION, N_("FILE"), 0,
+   N_("exclude directories containing FILE"), GRID+1 }, 
   {"no-recursion", NO_RECURSION_OPTION, 0, 0,
    N_("avoid descending automatically in directories"), GRID+1 },
   {"one-file-system", ONE_FILE_SYSTEM_OPTION, 0, 0,
@@ -953,7 +957,8 @@ enum read_file_list_state  /* Result of reading file name from the list file */
   {
     file_list_success,     /* OK, name read successfully */
     file_list_end,         /* End of list file */
-    file_list_zero         /* Zero separator encountered where it should not */
+    file_list_zero,        /* Zero separator encountered where it should not */
+    file_list_skip         /* Empty (zero-length) entry encountered, skip it */
   };
 
 /* Read from FP a sequence of characters up to FILENAME_TERMINATOR and put them
@@ -971,13 +976,15 @@ read_name_from_file (FILE *fp, struct obstack *stk)
 	{
 	  /* We have read a zero separator. The file possibly is
 	     zero-separated */
-	  /* FATAL_ERROR((0, 0, N_("file name contains null character"))); */
 	  return file_list_zero;
 	}
       obstack_1grow (stk, c);
       counter++;
     }
 
+  if (counter == 0 && c != EOF)
+    return file_list_skip;
+
   obstack_1grow (stk, 0);
 
   return (counter == 0 && c == EOF) ? file_list_end : file_list_success;
@@ -1058,31 +1065,42 @@ update_argv (const char *filename, struct argp_state *state)
 	open_fatal (filename);
     }
 
-  while ((read_state = read_name_from_file (fp, &argv_stk)) == file_list_success)
-    count++;
-
-  if (read_state == file_list_zero)
+  while ((read_state = read_name_from_file (fp, &argv_stk)) != file_list_end)
     {
-      size_t size;
+      switch (read_state)
+	{
+	case file_list_success:
+	  count++;
+	  break;
 
-      WARN ((0, 0, N_("%s: file name read contains nul character"),
-	     quotearg_colon (filename)));
+	case file_list_end: /* won't happen, just to pacify gcc */
+	  break;
 
-      /* Prepare new stack contents */
-      size = obstack_object_size (&argv_stk);
-      p = obstack_finish (&argv_stk);
-      for (; size > 0; size--, p++)
-	if (*p)
-	  obstack_1grow (&argv_stk, *p);
-        else
-	  obstack_1grow (&argv_stk, '\n');
-      obstack_1grow (&argv_stk, 0);
-      count = 1;
+	case file_list_zero:
+	  {
+	    size_t size;
 
-      /* Read rest of files using new filename terminator */
-      filename_terminator = 0;
-      while (read_name_from_file (fp, &argv_stk) == file_list_success)
-	count++;
+	    WARN ((0, 0, N_("%s: file name read contains nul character"),
+		   quotearg_colon (filename)));
+
+	    /* Prepare new stack contents */
+	    size = obstack_object_size (&argv_stk);
+	    p = obstack_finish (&argv_stk);
+	    for (; size > 0; size--, p++)
+	      if (*p)
+		obstack_1grow (&argv_stk, *p);
+	      else
+		obstack_1grow (&argv_stk, '\n');
+	    obstack_1grow (&argv_stk, 0);
+	    count = 1;
+	    /* Read rest of files using new filename terminator */
+	    filename_terminator = 0;
+	    break;
+	  }
+
+	case file_list_skip:
+	  break;
+	}
     }
 
   if (!is_stdin)
@@ -1492,6 +1510,10 @@ parse_opt (int key, char *arg, struct argp_state *state)
       exclude_caches_option = true;
       break;
 
+    case EXCLUDE_TAG_OPTION:
+      add_exclude_tag (arg);
+      break;
+      
     case FORCE_LOCAL_OPTION:
       force_local_option = true;
       break;
@@ -2332,7 +2354,7 @@ main (int argc, char **argv)
 
   if (stdlis == stdout)
     close_stdout ();
-  else if (ferror (stderr) || fclose (stderr) != 0) 
+  else if (ferror (stderr) || fclose (stderr) != 0)
     exit_status = TAREXIT_FAILURE;
 
   return exit_status;
