@@ -264,7 +264,7 @@ xheader_format_name (struct tar_stat_info *st, const char *fmt, size_t n)
 	case 'f':
 	  if (st)
 	    {
-	      base = base_name (st->orig_file_name);
+	      base = last_component (st->orig_file_name);
 	      len += strlen (base) - 2;
 	    }
 	  break;
@@ -647,12 +647,10 @@ void
 xheader_read (union block *p, size_t size)
 {
   size_t j = 0;
-  size_t nblocks;
 
   free (extended_header.buffer);
   size += BLOCKSIZE;
   extended_header.size = size;
-  nblocks = (size + BLOCKSIZE - 1) / BLOCKSIZE;
   extended_header.buffer = xmalloc (size + 1);
   extended_header.buffer[size] = '\0';
 
@@ -1094,9 +1092,10 @@ ctime_decoder (struct tar_stat_info *st,
 
 static void
 mtime_coder (struct tar_stat_info const *st, char const *keyword,
-	     struct xheader *xhdr, void const *data __attribute__ ((unused)))
+	     struct xheader *xhdr, void const *data)
 {
-  code_time (st->mtime, keyword, xhdr);
+  const struct timespec mtime = data ? *(struct timespec *) data : st->mtime;
+  code_time (mtime, keyword, xhdr);
 }
 
 static void
